function [path] = astar_voronoi2(adj, X, Y, node_start, node_goal)
% Input: Adjacency matrix for graph G, X and Y coordinates of nodes,
% node_start, node_goal
% Output: List of nodes in order of shortest route

% f(n) = g(n) + h(n)
% g(n) cost of the path from from the start node to n
% h(n) heuristic function that estimates cost the cost of cheapest path
% from n to goal


% Initialize open and closed list
open_list = {};
closed_list = {};
path_struct = [];
open_list(1,:) = node_start;
path = node_start{6};
% f, g, h, x, y, node_nr i adjacency
while (isempty(open_list) ~= 1)
    % Get the current node by choosing min value
    [f_value,i] = min([open_list{:,1}]);
    node_current = open_list(i,:); 
    % Remove current node from open list
    open_list(i,:) = [];
    

    % Found the goal
    if node_current{4} == node_goal{4} && node_current{5} == node_goal{5}
        disp('Goal node found')
        iterator = node_goal{6};
        path = [];
        while iterator ~= node_start{6}
            index = find([path_struct(:,1)] == iterator);
            parent = path_struct(index,2);
            path(size(path,1)+1,:) = iterator;
            iterator = parent;
        end
        path(size(path,1)+1,:) = node_start{6};
        path = flipud(path);
        return
    end
    
    %Find set of neighbours
    [~,col] = find(adj(node_current{6},:));
    for jj=1:length(col)
        %disp(col(jj))
        pyt = (node_current{4}-X(col(jj)))^2+(node_current{5}-Y(col(jj)))^2;
        successor_current_cost = node_current{2} + sqrt(pyt);
        % Searches for node in lists, indexes in cell array
        indexOpen = find([open_list{:,6}] == col(jj));
        if (isempty(closed_list) ~= 1)
            indexClosed = find([closed_list{:,6}] == col(jj));
        end
        
        % Child is in open list
        if (isempty(indexOpen) ~= 1)
            if (open_list{indexOpen,2} <= successor_current_cost)
                continue;
            end
            open_list{indexOpen,2} = successor_current_cost;

        % Child is in closed list
        elseif (isempty(closed_list) ~= 1 && isempty(indexClosed) ~= 1)
            if (closed_list{indexClosed,2} <= successor_current_cost)
                continue;
            end
            h = sqrt((node_goal{4}-X(col(jj)))^2+(node_goal{5}-Y(col(jj)))^2);
            f = successor_current_cost+h;
            open_list(size(open_list,1)+1,:) = {f,successor_current_cost...
                ,h,X(col(jj)),Y(col(jj)),col(jj)};
            closed_list(indexClosed,:) = [];
        else
            h = sqrt((node_goal{4}-X(col(jj)))^2+(node_goal{5}-Y(col(jj)))^2);
            f = successor_current_cost+h;
            open_list(size(open_list,1)+1,:) = {f,successor_current_cost...
                ,h,X(col(jj)),Y(col(jj)),col(jj)};
        end
        
        % Add to path
        if (isempty(path_struct))
            % node, parent
            path_struct(1,:) = [col(jj), node_current{6}];
        else
            indexPath= find([path_struct(:,1)] == col(jj));
            if ~isempty(indexPath)
                path_struct(indexPath,:) = [col(jj), node_current{6}];
            else
                path_struct(size(path_struct,1)+1,:) = [col(jj), node_current{6}];
            end
            
        end
    end
    % Add current node to closed list
    closed_list(size(closed_list,1)+1,:) = node_current;
end


disp('Error: goal node not found');
end
