% Global path planning

% This script generates vertices and adjacency matrix for obstacle courses


% Create voronoi
% Define map and obstacles

% --------------------------------------------------------------
obstacle_1x = [10 8.6 11 15 21 23 22 17];
obstacle_1y = [10 14.7 21.11 23.91 23.40 19.07 15 9.55];
obstacle_2x = [30.8 40 40 39 35 33 30];
obstacle_2y = [50 50 45 38 35 38 44];
obstacle_3x = [37 41 45 44 40 35 32 35];
obstacle_3y = [17 18 16 10 6 5 8.7 14];
obstacle_4x = [32 33 36 35 33];
obstacle_4y = [26 28 27 24.7 24.6];
obstacle_5x = [10 15 18 15 10 6.5];
obstacle_5y = [44 44 39 35 35 39.6];
obstacles_x = [obstacle_1x NaN obstacle_2x NaN obstacle_3x NaN obstacle_4x NaN obstacle_5x];
obstacles_y = [obstacle_1y NaN obstacle_2y NaN obstacle_3y NaN obstacle_4y NaN obstacle_5y];

borders_x = [0 0 0 0 0 0 0 0 0 0 0 5 10 15 20 25 30 35 40 45 50 50 50 50 50 50 50 50 50 50 50 45 40 35 30 25 20 15 10 5];
borders_y = [0 5 10 15 20 25 30 35 40 45 50 50 50 50 50 50 50 50 50 50 50 45 40 35 30 25 20 15 10 5 0 0 0 0 0 0 0 0 0 0];

% Generator points Voronoi
X = [obstacle_1x obstacle_2x obstacle_3x obstacle_4x obstacle_5x borders_x]; %obstacles
Y = [obstacle_1y obstacle_2y obstacle_3y obstacle_4y obstacle_5y borders_y]; %obstacles

% --------------------------------------------------------------
% Create Voronoi diagram
[vx,vy] = voronoi(X,Y);
plot(X,Y,'r+',vx,vy,'b-')
axis equal
hold on

patch(obstacle_1x',obstacle_1y','black');
patch(obstacle_2x',obstacle_2y','black');
patch(obstacle_3x',obstacle_3y','black');
patch(obstacle_4x',obstacle_4y','black');
patch(obstacle_5x',obstacle_5y','black');

x1 = vx(1,:);
x2 = vx(2,:);
y1 = vy(1,:);
y2 = vy(2,:);

%nodes = [x1 x2; y1 y2];

% Compute laplacian
E=[x1' y1' x2' y2'];
[A,XY]= unmesh(E);
adj = full(A);

% convert laplacian to adjacency matrix
for i=1:size(adj,1)
    for j = 1:size(adj,2)
        if (i == j)
            adj(i,j) = 0;
        else
            if(adj(i,j) == -1)
                adj(i,j) = 1;
                adj(j,i) = 1;
            end
        end
    end
end


% remove waypoints outside map borders
for i=1:size(XY,1)
    if (XY(i,1) < 0 || XY(i,1) > 50 || XY(i,2) < 0 || XY(i,2) > 50)
        adj(i,:) = 0;
        adj(:,i) = 0;
    end
end

% remove waypoints inside or on the edge of obstacles
in = inpolygons(XY(:,1),XY(:,2),obstacles_x,obstacles_y);
adj(in,:) = 0;
adj(:,in) = 0;

%scatter(XY(~in,1), XY(~in,2),20,[0 0 0],'filled')
% axis equal;
xlim([0 50]);
ylim([0 50]);
[X_g, Y_g] = gplot(adj,XY);
plot(X_g, Y_g, 'Color', [0 1 1]);

% Insert start and goal node in roadmap
n_nodes = size(adj,1);
start_x = 0;
start_y = 0;
goal_x = 50;
goal_y = 50;

% Connect start and goal nodes to roadmap
XY(n_nodes+1,1) = start_x;
XY(n_nodes+1,2) = start_y;
XY(n_nodes+2,1) = goal_x;
XY(n_nodes+2,2) = goal_y;

distances = zeros(10,2);

for i = 1:n_nodes
    distances(i,1) = sqrt((XY(i,1)-start_x)^2+(XY(i,2)-start_y)^2);
    distances(i,2) = sqrt((XY(i,1)-goal_x)^2+(XY(i,2)-goal_y)^2);
end

[~, closest_nodes_start] = sort(distances(:,1),'ascend');
[~, closest_nodes_goal] = sort(distances(:,2),'ascend');

% Connect to 3 closest neighbours

for i = 1:1
    % indeks i adj: closest_nodes_goal(i)
    adj(n_nodes+1, closest_nodes_start(i)) = 1;
    adj(closest_nodes_start(i), n_nodes+1) = 1;
    
    adj(n_nodes+2, closest_nodes_goal(i)) = 1;
    adj(closest_nodes_goal(i), n_nodes+2) = 1;
end

[X_g, Y_g] = gplot(adj,XY);
plot(X_g, Y_g, 'Color', [0 0 1]);

% Define nodes for A* search
node_start = {sqrt((goal_x-start_x)^2+(goal_y-start_y)^2),0,...
    sqrt((goal_x-start_x)^2+(goal_y-start_y)^2), start_x, start_y, n_nodes+1};
node_goal = {0,0,0,goal_x, goal_y, n_nodes+2};

% Retrieve path
wayPoints = getPath(adj, XY(:,1), XY(:,2), node_start, node_goal);

% Plot path
plot(wayPoints(:,1), wayPoints(:,2));


hold off