function [sys,ThrusterPos,str,ts] = thrustallocationPseudo(t,x,u,flag,Thruster,C,constraints) 
% Function for finding the optimal path to desired angle
% Work in progress  
% Author: Preben Frederich
% 12/12-2015

% Version 1.0 : Pseudo inverse optimalization created
%               12/12-2015 Preben Frederich
% Version 1.1 : Desired angle implemented
%               02/03-2016 Preben Frederich
% Version 1.2 : Individual & coupled constrained thrust allocation implemented
%               07/04-2016 Preben Frederih
% Version 1.3 : Update
%               18/03-2019 Jon Bjrn

switch flag

  % Initialization
  case 0
    [sys,ThrusterPos,str,ts]=mdlInitializeSizes(Thruster);

  % Output
  case 3
    sys=mdlOutputs(t,x,u,Thruster);

  % Update  
  case 2
    sys=mdlUpdate(t,x,u,Thruster,C,constraints); 
    
      case {1,4,}
    sys=[];

  % Case 9 not yet implemented. Used for terminate signal if errors where
  % to occure.
  case 9
      sys=mdlTerminate(t,x,u,Thruster);

  % Unexpected flags
  otherwise
    error(['Unhandled flag = ',num2str(flag)]);

end

function [sys,x0,str,ts]=mdlInitializeSizes(Thruster)
% Only used when starting the system


sizes = simsizes;  

sizes.NumContStates  = 0;   % Number of continuous states in the system, 
sizes.NumDiscStates  = 12;  % Number of discrete states in the system 
sizes.NumOutputs     = 12;  % Number of outputs (Output of each angle of each thruster and desired thrust of each thruster)
sizes.NumInputs      = 10;  % Number of inputs2 (Input: Desired angle and Desired thrust vector)
sizes.DirFeedthrough = 0;   % No element which is sent directly through the system
sizes.NumSampleTimes = 1;  
sys = simsizes(sizes); 


%% Determines initial angles on thrusters + initial thrust. (FORSK SITT INITIAL ANGLE INN SOM EGEN VARIABLE)
x0 = [Thruster.T1(3) Thruster.T2(3) Thruster.T3(3) Thruster.T4(3) Thruster.T5(3) Thruster.T6(3)  0 0 0 0 0 0]';

str = [];  

ts  = [-1 0];


function sys=mdlUpdate(t,x,u,Thruster,C,constraints)
% Complete system goes here!
ThrusterPosX = [Thruster.T1(1) Thruster.T2(1) Thruster.T3(1) Thruster.T4(1) Thruster.T5(1) Thruster.T6(1)];%[x(1),x(2),x(3),x(4),x(5),x(6)]';
ThrusterPosY = [Thruster.T1(2) Thruster.T2(2) Thruster.T3(2) Thruster.T4(2) Thruster.T5(2) Thruster.T6(2)];%[x(7),x(8),x(9),x(10),x(11),x(12)]';

% Parameters thrusters
tc =  [u(1),u(2),u(3)]'; 
Alpha = [u(4),u(5),u(6),u(7),u(8),u(9)]';
epsilon = u(10);

%% Thrust allocation algorithm

n_t = length(ThrusterPosX); % number of thrusters
B_surge = zeros(n_t,1); % Shells for faster computing
B_sway = zeros(n_t,1);  % Shells for faster computing
B_yaw = zeros(n_t,1);   % Shells for faster computing





% Constraints for Thruster-Thruster interaction. Can be seen in "Initfile"
% how they are found
C1 = C(1,:);
C2 = C(2,:);
C3 = C(3,:);
C4 = C(4,:);
C5 = C(5,:);
C6 = C(6,:);

K = ones(1,6); 

if constraints
    % Ok 12.05/16
    if (Alpha(1) < C1(1)-0.1  && Alpha(1) > C1(2)+0.1) || (Alpha(1) > C1(3)+0.1 && Alpha(1) < C1(4)-0.1)
        K(1) = 0;
    else 
        K(1) = K(1);
    end
    % OK 12.05/16
    if (Alpha(2) < C2(1)-0.1  && Alpha(2) > C2(2)+0.1) || (Alpha(2) < C2(3)-0.1 && Alpha(2) > C2(4)+0.1)
        K(2) = 0;
    else 
        K(2) = K(2);
    end
    % OK 12.05/16
    if (Alpha(3) > C3(1)+0.1  && Alpha(3) < C3(2)-0.1) || (Alpha(3) > C3(3)+0.1 && Alpha(3) < C3(4)-0.1)
        K(3) = 0;
    else 
        K(3) = K(3);
    end
    % OK 12.05/16
    if (Alpha(4) > C4(1)+0.1  && Alpha(4) < C4(2)-0.1) || (Alpha(4) < C4(3)- 0.1 && Alpha(4) > C4(4)+0.1)
        K(4) = 0;
    else 
        K(4) = K(4);
    end
    % OK 12.05/16
    if (Alpha(5) > C5(1)+0.1  && Alpha(5) < C5(2)-0.1) || (Alpha(5) > C5(3)+0.1 && Alpha(5) < C5(4)-0.1)
        K(5) = 0;
    else 
        K(5) = K(5);
    end
    % OK 12.05/16
    if (Alpha(6) < C6(1)-0.1  && Alpha(6) > C6(2)+0.1) || (Alpha(6) < C6(3)-0.1 && Alpha(6) > C6(4)+0.1)
        K(6) = 0;
    else 
        K(6) = K(6);
    end

    K = [K(1) K(2) K(3) K(4) K(5) K(6)];
end

%% Finding Optimal angle from desired forces tau
ue = [1 0 1 0 1 0 1 0 1 0 1 0;
      0 1 0 1 0 1 0 1 0 1 0 1;
      ThrusterPosY(1) ThrusterPosX(1) ThrusterPosY(2) ThrusterPosX(2) ThrusterPosY(3) ThrusterPosX(3) ...
      ThrusterPosY(4) ThrusterPosX(4) ThrusterPosY(5) ThrusterPosX(5) ThrusterPosY(6) ThrusterPosX(6)];
[U_singular,S_singular,V_singular] = svd(ue,'econ');
S_singular_cross = 1./S_singular;
S_singular_cross(~isfinite(S_singular_cross))=0;
 if (S_singular_cross(3,3)>2*S_singular_cross(2,2)) 
     S_singular_cross(3,3)=0;
 end
u0 = V_singular*S_singular_cross*U_singular'*tc;
Alpha = [atan2d(u0(2),u0(1)) atan2d(u0(4),u0(3)) atan2d(u0(6),u0(5)) atan2d(u0(8),u0(7)) atan2d(u0(10),u0(9)) atan2d(u0(12),u0(11))];

% Ensuring the angles are between [-180 180] degrees
for i=1:1:n_t
  if Alpha(i) > 180
      Alpha(i) = Alpha(i) - 360;
  elseif Alpha(i) < -180
      Alpha(i) = Alpha(i) + 360;
  else
      Alpha(i);
  end
end

for i = 1:1:n_t
  B_surge(i) = cosd(Alpha(i));     %Thrust in x-direction for cartesian
  B_sway(i) = sind(Alpha(i));      %Thrust in y-direction for cartesian
  B_yaw(i) = (sind(Alpha(i))*ThrusterPosX(i) - cosd(Alpha(i))*ThrusterPosY(i));
end

B = [B_surge,B_sway,B_yaw]';
B = B*diag(K);
[U_singular,S_singular,V_singular] = svd(B,'econ');
S_singular_cross = 1./S_singular;
S_singular_cross(~isfinite(S_singular_cross))=0;
 if (S_singular_cross(3,3)>2*S_singular_cross(2,2)) 
     S_singular_cross(3,3)=0;
 end

T_optimized = V_singular*S_singular_cross*U_singular';

u_d = T_optimized*tc+(tc(1)+tc(2)+tc(3))*epsilon;           % Optimized Thrust



%% Constraints
if constraints
    % Thruster 1 OK 13.05/16
    %C1 = [49.7 29.4 -151 -130 ];

    if (Alpha(1) < C1(1)  && Alpha(1) > C1(2)) %&& u_d(1) > 0) || (alpha1(1) < C1(1)  && alpha1(1) > C1(2) && u_d(1) < 0)
        c1 = [C1(1) C1(2)];
        [C_1,alpha_new1] = min(abs(c1-Alpha(1)));
        Alpha(1) = c1(alpha_new1);
    elseif (Alpha(1) > C1(3) && Alpha(1) < C1(4) && u_d(1) > 0) || (Alpha(1) > C1(3) && Alpha(1) < C1(4) && u_d(1) < 0)
        c1 = [C1(3) C1(4)];
        [C_1,alpha_new1] = min(abs(c1-Alpha(1)));
        Alpha(1) = c1(alpha_new1);
    else 
        Alpha(1) = Alpha(1);
    end
    if Alpha(1) < C1(1)-0.1  && Alpha(1) > C1(2)+0.1 || Alpha(1) > C1(3)+0.1 && Alpha(1) < C1(4)-0.1
        u_d(1) = 0;
    else 
        u_d(1) = u_d(1);
    end

    %% Thruster 2 OK 13.05/16
    %C2 = [99.85 80.15 50 29];
    if (Alpha(2) < C2(1)  && Alpha(2) > C2(2) && u_d(2) > 0) || (Alpha(2) < C2(1)  && Alpha(2) > C2(2) && u_d(2) < 0) 
        c2 = [C2(1) C2(2)];
        [C_2,alpha_new2] = min(abs(c2-Alpha(2)));
        Alpha(2) = c2(alpha_new2);
    elseif (Alpha(2) < C2(3) && Alpha(2) > C2(4) && u_d(2) > 0) || (Alpha(2) < C2(3) && Alpha(2) > C2(4) && u_d(2) < 0)
        c2 = [C2(3) C2(4)];
        [C_2,alpha_new2] = min(abs(c2-Alpha(2)));
        Alpha(2) = c2(alpha_new2);
    else 
        Alpha(2) = Alpha(2);
    end
    if (Alpha(2) < C2(1)-0.1  && Alpha(2) > C2(2)+0.1) || (Alpha(2) < C2(3)-0.1 && Alpha(2) > C2(4)+0.1)
        u_d(2) = 0;
    else 
        u_d(2) = u_d(2);
    end

    %% Thruster 3 OK 13.05/16
    % C3 = [-99.85 -80.15 -50 -29]; 

    if (Alpha(3) > C3(1)  && Alpha(3) < C3(2) && u_d(3) > 0) || (Alpha(3) > C3(1)  && Alpha(3) < C3(2) && u_d(3) < 0)
        c3 = [C3(1) C3(2)];
        [C_3,alpha_new3] = min(abs(c3-Alpha(3)));
        Alpha(3) = c3(alpha_new3);
    elseif (Alpha(3) > C3(3) && Alpha(3) < C3(4) && u_d(3) > 0) || (Alpha(3) > C3(3) && Alpha(3) < C3(4) && u_d(3) < 0) 
        c3 = [C3(3) C3(4)];
        [C_3,alpha_new3] = min(abs(c3-Alpha(3)));
        Alpha(3) = c3(alpha_new3);
    else 
        Alpha(3) = Alpha(3);
    end
    if (Alpha(3) > C3(1) + 0.1  && Alpha(3) < C3(2) - 0.1) || (Alpha(3) > C3(3) + 0.1 && Alpha(3) < C3(4) - 0.1)
        u_d(3) = 0;
    else 
        u_d(3) = u_d(3);
    end

    %% Thruster 4 OK 13.05/16
    %C4 = [-52.8 -34.2 52.8 34.2]; 
    if (Alpha(4) > C4(1)  && Alpha(4) < C4(2) && u_d(4) > 0) || (Alpha(4) > C4(1)  && Alpha(4) < C4(2) && u_d(4) < 0)
        c4 = [C4(1) C4(2)];
        [C_4,alpha_new4] = min(abs(c4-Alpha(4)));
        Alpha(4) = c4(alpha_new4);
    elseif (Alpha(4) < C4(3) && Alpha(4) > C4(4) && u_d(4) > 0) || (Alpha(4) < C4(3) && Alpha(4) > C4(4) && u_d(4) < 0) 
        c4 = [C4(3) C4(4)];
        [C_4,alpha_new4] = min(abs(c4-Alpha(4)));
        Alpha(4) = c4(alpha_new4);
    else 
        Alpha(4) = Alpha(4);
    end
    if (Alpha(4) > C4(1)+0.1  && Alpha(4) < C4(2)-0.1) || (Alpha(4) < C4(3)-0.1 && Alpha(4) > C4(4)+0.1)
        u_d(4) = 0;
    else 
        u_d(4) = u_d(4);
    end

    %% Thruster 5 OK 13.05/16
    %C5 = [-98.25 -81.75 -145.8 -127.2];
    if (Alpha(5) > C5(1)  && Alpha(5) < C5(2) && u_d(5) > 0) || (Alpha(5) > C5(1)  && Alpha(5) < C5(2) && u_d(5) < 0)
        c5 = [C5(1) C5(2)];
        [C_5,alpha_new5] = min(abs(c5-Alpha(5)));
        Alpha(5) = c5(alpha_new5);
    elseif (Alpha(5) > C5(3) && Alpha(5) < C5(4) && u_d(5) > 0)  || (Alpha(5) > C5(3) && Alpha(5) < C5(4) && u_d(5) < 0)
        c5 = [C5(3) C5(4)];
        [C_5,alpha_new5] = min(abs(c5-Alpha(5)));
        Alpha(5) = c5(alpha_new5);
    else 
        Alpha(5) = Alpha(5);
    end
    if (Alpha(5) > C5(1)+0.1  && Alpha(5) < C5(2)-0.1) || (Alpha(5) > C5(3)+0.1 && Alpha(5) < C5(4)-0.1)
        u_d(5) = 0;
    else 
        u_d(5) = u_d(5);
    end

    %% Thruster 6 OK 13.05/16
    %C6 = [98.75 81.75 145.8 127.2];
    if (Alpha(6) < C6(1)  && Alpha(6) > C6(2) && u_d(6) > 0) || (Alpha(6) < C6(1)  && Alpha(6) > C6(2) && u_d(6) < 0)
        c6 = [C6(1) C6(2)];
        [C_6,alpha_new6] = min(abs(c6-Alpha(6)));
        Alpha(6) = c6(alpha_new6);
    elseif (Alpha(6) < C6(3) && Alpha(6) > C6(4) && u_d(6) > 0) || (Alpha(6) < C6(3) && Alpha(6) > C6(4) && u_d(6) < 0)
        c6 = [C6(3) C6(4)];
        [C_6,alpha_new6] = min(abs(c6-Alpha(6)));
        Alpha(6) = c6(alpha_new6);
    else 
        Alpha(6) = Alpha(6);
    end
    if (Alpha(6) < C6(1)-0.1  && Alpha(6) > C6(2)+0.1) || (Alpha(6) < C6(3)-0.1 && Alpha(6) > C6(4)+0.1)
        u_d(6) = 0;
    else 
        u_d(6) = u_d(6);
    end
    %% Coupled constraints: 
    % Thruster 2&3 
    if Alpha(2) == C2(2)
        Alpha(3) = Alpha(2) - (C2(2) - Alpha(2));
    elseif Alpha(2) == C2(1)
        Alpha(3) = Alpha(2) - (C2(1) - Alpha(2));
    elseif Alpha(2) == C2(3)
        Alpha(3) = Alpha(2) - (C2(3) - Alpha(2));
    elseif Alpha(2) == C2(4)
        Alpha(3) = Alpha(2) - (C2(4) - Alpha(2));

    elseif Alpha(3) == C3(2)
        Alpha(2) = Alpha(3) - (C3(2) - Alpha(3));
    elseif Alpha(3) == C3(1)
        Alpha(2) = Alpha(3) - (C3(1) - Alpha(3));
    elseif Alpha(3) == C3(3)
        Alpha(2) = Alpha(3) - (C3(3) - Alpha(3));
    elseif Alpha(3) == C3(4)
        Alpha(2) = Alpha(3) - (C3(4) - Alpha(3));
    end

    % Thruster 5&6
    if Alpha(5) == C5(2)
        Alpha(6) = Alpha(5) - (C5(2) - Alpha(5));
    elseif Alpha(5) == C5(1)
        Alpha(6) = Alpha(5) - (C5(1) - Alpha(5));
    elseif Alpha(5) == C5(3)
        Alpha(6) = Alpha(5) - (C5(3) - Alpha(5));
    elseif Alpha(5) == C5(4)
        Alpha(6) = Alpha(5) - (C5(4) - Alpha(5));

    elseif Alpha(6) == C6(2)
        Alpha(5) = Alpha(6) - (C6(2) - Alpha(6));
    elseif Alpha(6) == C6(1)
        Alpha(5) = Alpha(6) - (C6(1) - Alpha(6));
    elseif Alpha(6) == C6(3)
        Alpha(5) = Alpha(6) - (C6(3) - Alpha(6));
    elseif Alpha(6) == C6(4)
        Alpha(5) = Alpha(6) - (C6(4) - Alpha(6));
    end
    % Thruster 4 & 1
    if     Alpha(4) == C4(2)
        Alpha(1) = Alpha(4) - (C4(2) - Alpha(4));
    elseif Alpha(4) == C4(1)
        Alpha(1) = Alpha(4) - (C4(1) - Alpha(4));
    elseif Alpha(4) == C4(3)
        Alpha(1) = Alpha(4) - (C4(3) - Alpha(4));
    elseif Alpha(4) == C4(4)
        Alpha(1) = Alpha(4) - (C4(4) - Alpha(4));
    end

    if     Alpha(1) == C1(2)
        Alpha(4) = Alpha(1) - (C1(2) - Alpha(1));
    elseif Alpha(1) == C1(1)
        Alpha(4) = Alpha(1) - (C1(1) - Alpha(1));
    elseif Alpha(1) == C1(3)
        Alpha(4) = Alpha(1) - (C1(3) - Alpha(1));
    elseif Alpha(1) == C1(4)
        Alpha(4) = Alpha(1) - (C1(4) - Alpha(1));
    end
end 

Alpha = [Alpha(1) Alpha(2) Alpha(3) Alpha(4) Alpha(5) Alpha(6)]';    

Desired_thrust = [u_d(1) u_d(2) u_d(3) u_d(4) u_d(5) u_d(6)]';
Desired_thrust(Desired_thrust<0) = 0;

%% Update outputs
sys = [Alpha',Desired_thrust'];

function sys=mdlOutputs(t,x,u,Thruster)
% Giving the desired angle output [7 8 9 10 11 12][ 1 2 3 4 5 6] together with desired
% thrust [7 8 9 10 11 12]
sys=[x(1),x(2),x(3),x(4),x(5),x(6),x(7),x(8),x(9),x(10),x(11),x(12)];

function sys=mdlTerminate(t,x,u,Thruster) 
% Not yet implemented
sys = [];